=begin 
TIG (c) 2012-2013
All Rights Reserved.
THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

  TIG-Smart_offset.rb

Offsets edges of a specified face by a distance 'smartly'

  TIG::Smart_offset.new(face, distance)

Overview:
A +ve distance offsets the face's outer_loop outside of the loop.
A -ve distance offsets  the face's outer_loop inside the loop.
If the face's outer_loop perimeter contains a thin inlet or thin peninsula 
then twisted loops are avoided and trimmed off, only a single non-twisted 
new face and edge-set is created.
New edges crossing other coplanar faces will split them as appropriate.
Unlike the native 'face.offset', with an internal [-ve distance] any internal 
'holes' in the face or its connected faces are preserved and any new 
offet edges crossing them are left 'unfaced', unless the final [optional] 
argument is set 'true', i.e.
  TIG::Smart_offset.new(face, distance, true)
then any such 'unfaced' edges crossing 'holes' are deleted.  
Holes in unconnected but coplanar faces that overlap with the new geometry 
might be 'infilled', just as with the native 'offset'.
On completion it returns an array containing the original face and any new 
face[s] created or affected by the creation of the new geometry, or 'nil' 
if no face could be created - e.g. the face was too small to contain the 
internal offset loop using the given -ve distance.
The new face[s] have the same front/back material and layer as the face
itself.
Some very convoluted faces with 're-entrant' forms and 'inlets' can cause 
failures, because a fully 'neatened' set of offset edges cannot be determined...

Usage:
For developers it can be used as a method to call from within their own scripts.  
Add a "require 'TIG-Smart_offset.rb'" line into your code to ensure it's 
loading, OR adjust the [modified] code to run only within your own Module/Class.
Remember to distribute this file with your scripts, because users probably won't 
have it installed.

  f=Sketchup.active_model.selection[0]; TIG::Smart_offset.new(f, 1.2)

OR from the 'Tools' menu > 'TIG.Smart_offset'.
OR from the Context-menu > 'TIG.Smart_offset' .
OR from the [activated] Toolbar > 'TIG.Smart_offset'.

Which use the method:
  TIG::Smart_offsetter.dialog()

Enter the required offset in current units.
Or add a units suffix for non-current dimensions.
A +ve offset is applied outside of the face's outer_loop and a -ve offset is 
made inside.
It mimicks the native offset-tool in most respects, other than removing 
spurious twisted offset line/loops.
If you use the tool this way then ALL selected faces are offset as specified.
The last used offset is remembered during that session.
Some faces might not be offset-able, e.g. inward offset > width/2 !
It is one step undo-able for all offsets.

Version:
1.0 20121221 First issue.
1.1 20121221 Made more robust.

Donations:
By PayPal.com to info @ revitrev.org
###
=end

require 'sketchup.rb'

module TIG

  module Smart_offsetter
  
  	def self.suitable?()
		faces=Sketchup.active_model.selection.grep(Sketchup::Face)
		if faces[0]
			return true
		else
			return false
		end
	end#def
		
	def self.dialog()
		model=Sketchup.active_model
		ss=model.selection
		faces=ss.grep(Sketchup::Face)
		unless faces[0]
			UI.messagebox("TIG.Smart_offset:\n\nPreSelect Face[s] !")
			return nil
		end
		@dist = 0.0.mm unless @dist
		prompts=["Offset [+/-ve=out/inside]: "]
		values=[@dist]
		pops=[""]
		title="TIG.Smart_offset"
		results=inputbox(prompts, values, pops, title)
		return nil unless results
		@dist, = results
		if @dist==0
			UI.messagebox("TIG.Smart_offset:\n\nOffset of ZERO does nothing !")
			self.dialog()
			return nil
		end
		tidy=false
		#ss.clear
		model.start_operation("TIG.Smart_offset")
		faces.each{|face| TIG::Smart_offset.new(face, @dist, tidy) }
		model.commit_operation
	end#def
	
	unless file_loaded?(__FILE__)
		dir=File.join(File.dirname(__FILE__), "TIG-Smart_offset")
		cmd=UI::Command.new('TIG.Smart_offset'){self.dialog()}
		cmd.set_validation_proc{
			if self.suitable?()
				MF_ENABLED
			else
				MF_DISABLED
				MF_GRAYED
			end
		}
		cmd.tooltip='TIG.Smart_offset'
		cmd.status_bar_text='TIG.Smart_offset: Offsets Select Faces, Smartly...'
		UI.menu('Tools').add_item(cmd)
		cmd.small_icon=File.join(dir, 'TIG-Smart_offset.png')
		cmd.large_icon=File.join(dir, 'TIG-Smart_offset.png')
		toolbar=UI::Toolbar.new('TIG.Smart_offset')
		toolbar.add_item(cmd)
		toolbar.restore if toolbar.get_last_state.abs==1 #TB_VISIBLE/NEVER
		UI.add_context_menu_handler{|menu|
		    menu.add_item(cmd) if self.suitable?()
		}
	end#unless
	file_loaded(__FILE__)
	
  end#module

  class Smart_offset
  
	def initialize(face=nil, dist=nil, tidy=false)
      begin
		return nil unless face && face.is_a?(Sketchup::Face) && face.valid?
		return nil unless dist && ((dist.class==Fixnum || dist.class==Float || dist.class==Length) && dist!=0)
        model=Sketchup.active_model
        norm=face.normal
        mat=face.material
        bat=face.back_material
        lay=face.layer
        ents=face.parent.entities
        afaces=[]
        allcon=face.all_connected
        allcon.find_all{|e|
          e.class==Sketchup::Face
        }.each{|f|
          next unless f.valid?
          afaces << [f, f.normal, f.material, f.back_material, f.layer]
        }
        selected=false
        ss=Sketchup.active_model.selection
        selected=true if ss.to_a.include?(face)
        ###
        loop=face.outer_loop
		verts=loop.vertices
		pts=[]
        ots=[]
		### create array pts of offset points from face 
		verts.each_index{|a|
            pta=verts[a].position
            vec1=pta.vector_to(verts[a-(verts.length-1)].position).normalize
			vec2=pta.vector_to(verts[a-1].position).normalize
            ang=vec1.angle_between(vec2)/2.0
			vec3=(vec1+vec2).normalize
		    if vec1.parallel?(vec2)
              ang=90.degrees 
              tpa=pta.offset(vec1)
              tr=Geom::Transformation.rotation(pta, norm, ang)
              tpa.transform!(tr)
              vec3=pta.vector_to(tpa)
            end
			if vec3.valid? and vec3.length>0
                vec3.length=0.25.mm ### v1.1
                tpt=pta.offset(vec3)
                if dist > 0 ### ensure start at internal corner is extl.
                  if face.classify_point(tpt)==Sketchup::Face::PointInside
                    vec3.reverse!
                  end
                else ### < 0 should be intl.
                  if face.classify_point(tpt)!=Sketchup::Face::PointInside
                    vec3.reverse!
                  end
                end
                vec3.length=(dist/Math::sin(ang)).abs
                ots << pta.to_a
				pts << pta.offset(vec3).to_a
			end#if
		}
        pts.uniq!
        ###
        unless pts[2]
          puts "#{face} can't offset !"
          p pts
          gp.erase! if gp.valid?
          return nil
        end
		### get vectors
        ovs=[]
        ots.each_with_index{|p, i|
          if i==ots.length-1
            ovs << p.vector_to(ots[0])
          else
            ovs << p.vector_to(ots[i+1])
          end
        }
        ### make set of nested groups and add edges
        gp=ents.add_group()
        gents=gp.entities
        ggp=gents.add_group()
        ggents=ggp.entities
        gggp=ggents.add_group()
        gggents=gggp.entities
        ggggp=gggents.add_group()
        ggggents=ggggp.entities
        ### add edges
        edges=ggggents.add_edges(pts<<pts[0])
        ###
        ggggp.explode ### to force geom to merge
        ### reversed edges 
        redges=[]
        ovs.each_with_index{|v, i|
          next unless edges[i].valid?
          redges << edges[i] if edges[i].line[1].normalize != v.normalize
        }
        redges.uniq!
        redges.each{|e|
          next unless e.valid?
          vs=e.start
          ve=e.end
          ps=vs.position
          pe=ve.position
          li=e.line
          vs.edges
          ve.edges
          es=(vs.edges-[e])[0]
          next if not es
          ee=(ve.edges-[e])[0]
          next if not ee
          vecs=es.line[1]
          vece=ee.line[1]
          e.erase!
          rayt=model.raytest([ps, vecs])
          if rayt and rayt[1].include?(gggp)
            v=ps.vector_to(rayt[0])
            tr=Geom::Transformation.translation(v)
            gggents.transform_entities(tr, vs)
          end
          rayt=model.raytest([pe, vece.reverse])
          if rayt and rayt[1].include?(gggp)
            v=pe.vector_to(rayt[0])
            tr=Geom::Transformation.translation(v)
            gggents.transform_entities(tr, ve)
          end
        }
        ###
        gggp.explode ### to force geom to merge
        ###
        tr=Geom::Transformation.new()
        ggents.intersect_with(true, tr, ggents, tr, true, loop.edges)
        edges=ggents.to_a.find_all{|e|e.class==Sketchup::Edge}
        teds=[]
        edges.each{|e|
          next unless e.valid?
          es=e.start.position
          teds << ggents.add_line(es, es.offset(norm,1))
        }
        ggents.intersect_with(true, tr, ggents, tr, true, ggents.to_a)
        ggp.explode ### to force geom to merge
        gents.erase_entities(teds)
        edges=gents.to_a.find_all{|e|e.class==Sketchup::Edge}
        togos=[]
        edges.each{|e|
          next unless e.valid?
          e.erase! if not e.start.edges[1] or not e.end.edges[1]
          next unless e.valid?
          if dist < 0 ### intl
            pt=e.start.position
            if face.classify_point(pt)==Sketchup::Face::PointOutside 
              togos << e
            elsif face.classify_point(pt)==Sketchup::Face::PointOnEdge 
              togos << e
            elsif face.classify_point(pt)==Sketchup::Face::PointOnVertex 
              togos << e
            end
            pt=e.end.position
            if face.classify_point(pt)==Sketchup::Face::PointOutside 
              togos << e
            elsif face.classify_point(pt)==Sketchup::Face::PointOnEdge 
              togos << e
            elsif face.classify_point(pt)==Sketchup::Face::PointOnVertex 
              togos << e
            end
          else ### >0 extl
            pt=e.start.position
            if face.classify_point(pt)==Sketchup::Face::PointInside 
              togos << e
            elsif face.classify_point(pt)==Sketchup::Face::PointOnEdge 
              togos << e
            elsif face.classify_point(pt)==Sketchup::Face::PointOnVertex 
              togos << e
            end
            pt=e.end.position
            if face.classify_point(pt)==Sketchup::Face::PointInside 
              togos << e
            elsif face.classify_point(pt)==Sketchup::Face::PointOnEdge 
              togos << e
            elsif face.classify_point(pt)==Sketchup::Face::PointOnVertex 
              togos << e
            end
          end
        }
        togos.uniq!
        togos.dup.each{|e|togos.delete(e) unless e.valid?}
        gents.erase_entities(togos)
        ###
        edges=gents.to_a.find_all{|e|e.class==Sketchup::Edge}
        edges.length.times{
          edges.each{|e|
            next unless e.valid?
            e.erase! unless e.start.edges[1] and e.end.edges[1]
          }
        }
        edges=gents.to_a.find_all{|e|e.class==Sketchup::Edge}
        edges.each{|e|
          next unless e.valid?
          e.find_faces
        }# unless e.faces[0]}
        edges=gents.to_a.find_all{|e|e.class==Sketchup::Edge}
        togos=[]
        edges.each{|e|
          next unless e.valid?
          togos << e unless e.faces.length==1
        }
        gents.erase_entities(togos)
        edges=gents.to_a.find_all{|e|e.class==Sketchup::Edge}
        edges.each{|e|
          e.erase! if e.valid? and (e.faces[1] or not e.faces[0])
        }
        facs=gents.find_all{|e|e.class==Sketchup::Face}
        if dist < 0 #|| loop.edges[0].faces[1]
          ### internal so needs NO faces #|| external with a 2nd face
          gents.erase_entities(facs) if facs[0]
        else ### external offset in free space so need a 'donut'
          gp.layer=lay
          facs.each{|f|
            next unless f.valid?
            f.reverse! unless f.normal==norm
            f.material=mat
            f.back_material=bat
            f.layer=lay
          }
          newf=gents.add_face(ots)
          gents.intersect_with(true, tr, gents, tr, true, gents.to_a)
          newf.erase! if newf.valid?
          edges=gents.find_all{|e|e.class==Sketchup::Edge}
          togos=[]
          edges.each{|e|
            next unless e.valid?
            togos << e unless e.faces.length==1
          }
          gents.erase_entities(togos)
        end
        ###
        edges=gents.find_all{|e|e.class==Sketchup::Edge}
        ###
		if gents[0]
          nents=gp.explode
        else
          puts "#{face} cannot offset !"
          gp.erase! if gp.valid?
          ss.add(face) if selected && face.valid?
          return nil
        end#if
        afaces.each{|a|
          next unless a[0].valid?
          a[0].reverse! unless a[0].normal== a[1]
          a[0].material=a[2]
          a[0].back_material=a[3]
          a[0].layer=a[4]
        }
        if tidy && (dist < 0 || loop.edges[0].faces[1])
          nents.find_all{|e|
            e.class==Sketchup::Edge
          }.each{|ee|
            next unless ee.valid?
            ee.erase! unless ee.faces[0]
          }
        elsif dist > 0 && loop.edges[0].faces[1]
          nents.find_all{|e|
            e.class==Sketchup::Edge
          }.each{|ee|
            next unless ee.valid?
            ee.find_faces
          }# unless f.faces[0]}
        end
        ###
        nfacs=[]
        edges.each{|e|
          next unless e.valid?
          nfacs << e.faces
        } if dist < 0
        allcon=face.all_connected
        ents.intersect_with(true, tr, ents, tr, true, allcon)
        nfacs.flatten!
        nfacs.uniq!
        if not nfacs[1] and dist < 0 ### v1.1
          puts "#{face} can not offset !"
		  p nfacs ### v1.1
		  p dist  ### v1.1
          ss.add(face) if selected && face.valid?
	      return nil
        end
        nfacs.each{|f|
          next unless f.valid?
          f.material=mat
          f.back_material=bat
          f.layer=lay
        }
        nfacs=[face]+(nfacs-[face]) ### 1st is the 'face'
        ss.add(face) if selected && face.valid?
      rescue Exception => err
        puts err
	    puts "#{face} can  not  offset !!"
        gp.erase! if gp.valid?
        ss.add(face) if selected && face.valid?
	    return nil
      end
      return nfacs
	end#def
	
  end#Smart_offset
  
end#TIG
